/*
 * Copyright (c) 2005-2010 KOM – Multimedia Communications Lab
 *
 * This file is part of Simonstrator.KOM.
 * 
 * Simonstrator.KOM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * 
 * PeerfactSim.KOM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with PeerfactSim.KOM.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


package de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.jam;

import java.util.Arrays;

import de.tudarmstadt.maki.simonstrator.api.Time;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.EnvironmentProperty;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.NumericVectoralProperty;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.ProbabilityDistribution;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.TemporalDependencyMatrix;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.location.Location;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.RoadNetworkEdge;

/**
 * @author Tobias Meuser (tobias.meuser@kom.tu-darmstadt.de)
 * @version 1.0 at 12.04.2018
 *
 */
public class VectoralJamProperty extends NumericVectoralProperty {
	public static final double SCALING = 6;
	private static final int DIMENSIONS = 15;

	// private static Map<Double, Double> _requiredStandardDeviation = new
	// HashMap<>();
	private static TemporalDependencyMatrix _temporalDependencyMatrix;

	public VectoralJamProperty(Location pLocation, RoadNetworkEdge pEdge) {
		super(pLocation, pEdge, DIMENSIONS);
	}

	/**
	 * 
	 */
	public VectoralJamProperty() {
		this(null, null);
	}

	public void setSpeed(double pSpeed) {
		double[] valueProbabilities = getValueProbabilities();
		Arrays.fill(valueProbabilities, 0);
		valueProbabilities[(int) (pSpeed / SCALING)] = 1;
	}

	public void setSpeed(double pSpeed, double pAccuracy, ProbabilityDistribution pDist) {
		switch (pDist) {
		case SIMPLE:
			throw new AssertionError("Currently not to be used!");
			// setSimple((int) (pSpeed / VectoralJamProperty.SCALING),
			// pAccuracy);
			// break;
		case GAUSSIAN:
			setGaussianWithAccuracy(pSpeed / VectoralJamProperty.SCALING, pAccuracy);
			break;
		default:
			throw new AssertionError("Unknown ProbablilityDistribution " + pDist);
		}
	}

	// private void setSimpleSpeed(double pSpeed, double pAccuracy) {
	// double[] valueProbabilities = getValueProbabilities();
	// Arrays.fill(valueProbabilities, 0);
	// int index = (int) (pSpeed / SCALING);
	// valueProbabilities[index] += pAccuracy;
	//
	// if (index + 1 < valueProbabilities.length) {
	// valueProbabilities[index + 1] += (1 - pAccuracy) / 2.0;
	// } else {
	// valueProbabilities[index - 1] += (1 - pAccuracy) / 2.0;
	// }
	//
	// if (index - 1 >= 1) {
	// valueProbabilities[index - 1] += (1 - pAccuracy) / 2.0;
	// } else {
	// valueProbabilities[index + 1] += (1 - pAccuracy) / 2.0;
	// }
	// }
	//
	// public void setGaussianSpeedWithAccuracy(double pSpeed, double pAccuracy)
	// {
	// double deviation = determineCorrectDeviation(pSpeed, pAccuracy);
	// setGaussianProbabilities(pSpeed / SCALING, deviation / SCALING);
	// }
	//
	// public void setGaussianSpeed(double pSpeed, double pDeviation) {
	// setGaussianProbabilities(pSpeed / SCALING, pDeviation / SCALING);
	// }

	@Override
	public VectoralJamProperty clone() {
		VectoralJamProperty vectoralJamProperty = new VectoralJamProperty(getLocation(), getEdge());
		vectoralJamProperty.setProbabilities(Arrays.copyOf(getValueProbabilities(), getValueProbabilities().length));
		vectoralJamProperty.setDetectionDate(getDetectionDate());
		return vectoralJamProperty;
	}

	@Override
	public VectoralJamProperty age(long pAge) {
		return (VectoralJamProperty) super.age(pAge);
	}

	@Override
	public EnvironmentProperty getDefaultProperty() {
		return new VectoralJamProperty(getLocation(), getEdge());
	}

	@Override
	public Double getValueAtIndex(int pIndex) {
		return pIndex * SCALING;
	}

	@Override
	public int getIndexForValue(Object pValue) {
		if (pValue instanceof Number) {
			return (int) (((Number) pValue).doubleValue() * SCALING);
		}
		throw new AssertionError("Unknown value!");
	}

	@Override
	public TemporalDependencyMatrix getDependencyMatrix() {
		if (_temporalDependencyMatrix == null) {
			TemporalDependencyMatrix temporalDependencyMatrix = new TemporalDependencyMatrix(DIMENSIONS);
			double TEMPORAL_CHANGE = 1 / ((double) JamInformationContainer.EVENT_DURATION / Time.SECOND) / 2.0;
			for (int x = 0; x < DIMENSIONS; x++) {
				double[] probabilities = new double[DIMENSIONS];
				if (x > 1) {
					probabilities[x - 1] += TEMPORAL_CHANGE;
				} else {
					probabilities[x + 1] += TEMPORAL_CHANGE;
				}
				if (x < DIMENSIONS - 1) {
					probabilities[x + 1] += TEMPORAL_CHANGE;
				} else {
					probabilities[x - 1] += TEMPORAL_CHANGE;
				}
				probabilities[x] += (1 - 2 * TEMPORAL_CHANGE);

				temporalDependencyMatrix.setTransitionProbability(x, probabilities);
			}
			_temporalDependencyMatrix = temporalDependencyMatrix;
		}
		return _temporalDependencyMatrix;
	}

	public double getProbabilityForValue(Object pValue) {
		if (pValue instanceof Double) {
			double value = (double) pValue;

			int index = (int) (value / SCALING);
			return getValueProbabilities()[index];
		}
		return 0;
	}

	/**
	 * @return
	 */
	public boolean isJammed() {
		if (getEdge() != null) {
			double originalMaxSpeed = getEdge().getOriginalMaxSpeed();

			double speed = getExpectation();

			if (speed < (((int) (originalMaxSpeed * 0.375 / VectoralJamProperty.SCALING))
					* VectoralJamProperty.SCALING)) {
				return true;
			}
		}
		return false;
	}
	
	@Override
	public int hashCode() {
		return Double.hashCode(getMostProbableValue());
	}

	@Override
	public boolean equals(Object pObj) {
		if (pObj instanceof VectoralJamProperty) {
			VectoralJamProperty property = (VectoralJamProperty) pObj;
			
			if (property.getMostProbableValue().equals(getMostProbableValue())) {
				return true;
			}

			return false;
		}
		return super.equals(pObj);
	}

	@Override
	public int getDefaultIndex() {
		return (int) (RoadNetworkEdge.getCorrespondingState(getEdge().getOriginalMaxSpeed()) / SCALING);
	}

	// public double determineCorrectDeviation(double speed, double pAccuracy) {
	// if (!_requiredStandardDeviation.containsKey(speed)) {
	// if (pAccuracy < 1) {
	// VectoralJamProperty vectoralJamProperty = new VectoralJamProperty(null,
	// null);
	//
	// double deviation = 10;
	//
	// boolean reduceChange = false;
	// double currentChange = deviation / 2;
	//
	// do {
	// vectoralJamProperty.setGaussianSpeed(speed, deviation);
	// if (vectoralJamProperty.getProbabilityForValue(speed) > pAccuracy) {
	// deviation += currentChange;
	// } else {
	// deviation -= currentChange;
	// reduceChange = true;
	// }
	//
	// if (reduceChange) {
	// currentChange /= 2.0;
	// }
	// } while (Math.round(vectoralJamProperty.getProbabilityForValue(speed) *
	// 10000) != Math
	// .round(pAccuracy * 10000));
	//
	// _requiredStandardDeviation.put(speed, deviation);
	// } else {
	// _requiredStandardDeviation.put(speed, 0d);
	// }
	// }
	//
	// return _requiredStandardDeviation.get(speed);
	// }
}